(*----------------------------------------------------------------------------*
 *  Direct3D sample for DirectX 8.x demostrating how to create X-File from    *
 *  your own internal data. As data source FaceData.pas by Jan Horn was taken *
 *  (from http://www.sulcaco.co.za)                                           *
 *                                                                            *
 *  Copyright (c) 2002 by Alexey Barkovoy                                     *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  Modified: 07-Apr-2002                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)

unit MainUnit;

{$I DirectX.inc}

interface

uses
  Windows, Messages, SysUtils, 
  D3DX8, D3DApp, D3DFont, D3DFile, D3DUtil, DXUtil, DXFile,
  {$I UseD3D8.inc};

type
  PMyVertex = ^TMyVertex;
  TMyVertex = packed record
    p: TD3DXVector3;
    n: TD3DXVector3;
  end;

const
  D3DFVF_VERTEX = D3DFVF_XYZ or D3DFVF_NORMAL;

type
//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the
//       generic functionality needed in all Direct3D samples. CMyD3DApplication
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_pFont: CD3DFont;          // Font for drawing text
    m_pObject: CD3DMesh;        // This is our mesh but already loaded from disk X-File
  protected
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;
    function ConfirmDevice(var pCaps: TD3DCaps8; dwBehavior: DWord; Format: TD3DFormat): HResult; override;

    procedure GenerateXFile; // this is the main procedure in example
  public
    constructor Create;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
  end;

implementation

uses FaceData;

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create(0);
  m_strWindowTitle      := 'X-File generator by Clootie';
  m_bUseDepthBuffer     := True;
  m_bShowCursorWhenFullscreen := True;

  m_pFont := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pObject := CD3DMesh.Create;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove;
var
  m: TD3DXMatrix;
begin
  D3DXMatrixTranslation(m, 0, 0, 1.5);
  m_pd3dDevice.SetTransform(D3DTS_VIEW, m);

  D3DXMatrixRotationY(m, -m_fTime);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, m);

  m_pd3dDevice.MultiplyTransform(D3DTS_WORLD,
    D3DXMatrixRotationX(m, -D3DX_PI*0.5)^);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render;
var
  matView: TD3DXMatrix;
begin
  // Clear the viewport
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
    $000000ff, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

    m_pd3dDevice.MultiplyTransform(D3DTS_WORLD,
      D3DXMatrixTranslation(matView, 0, -0.2, 0)^);

    // Render First face
    m_pObject.Render(m_pd3dDevice);

    m_pd3dDevice.MultiplyTransform(D3DTS_WORLD,
      D3DXMatrixTranslation(matView, 0, 0.4, 0)^);
    m_pd3dDevice.MultiplyTransform(D3DTS_WORLD,
      D3DXMatrixRotationZ(matView, D3DX_PI)^);

    // Render second face
    m_pObject.Render(m_pd3dDevice);

    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);
    m_pFont.DrawText(2, m_rcWindowClient.Bottom - 20,
      D3DCOLOR_ARGB(255,255,255,255),
      'Visit  "http://clootie.narod.ru"  for more demos...');

    // End the scene
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects;
begin
  // Restore the font
  m_pFont.InitDeviceObjects(m_pd3dDevice);

  // Let's do it
  GenerateXFile;

  // Load mesh genearated by us and show it...
  if FAILED(m_pObject.Create_(m_pd3dDevice, PChar('FaceText.x'))) then
  begin
    Result:= D3DAPPERR_MEDIANOTFOUND;
    Exit;
  end;
  m_pObject.SetFVF(m_pd3dDevice, D3DFVF_VERTEX);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects;
function D3DXColorFromDWord(c: DWord): TD3DXColor;
const
  f: Single = 1/255;
begin
  with Result do
  begin
    r:= f * Byte(c shr 16);
    g:= f * Byte(c shr  8);
    b:= f * Byte(c shr  0);
    a:= f * Byte(c shr 24);
  end;
end;
var
  matIdentity: TD3DXMatrix;
  matProj: TD3DXMatrix;
  fAspect: Single;
//  mtrl: TD3DMaterial8;
  light: TD3DLight8;
begin
  m_pFont.RestoreDeviceObjects;

  // Restore device-memory objects for the mesh
  m_pObject.RestoreDeviceObjects(m_pd3dDevice);

  // Set up the textures
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_CURRENT);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG2);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_CURRENT);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG2);

  m_pd3dDevice.SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_DISABLE);
  m_pd3dDevice.SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

  // Set miscellaneous render states
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE,   iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_SPECULARENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,        iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT,        $00444444);

  // Set the world matrix
  D3DXMatrixIdentity(matIdentity);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matIdentity);

  // Set the projection matrix
  fAspect:= m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

{  // Setup a material
  D3DUtil_InitMaterial(mtrl, 2.0, 2.0, 2.0, 1.0);
  m_pd3dDevice.SetMaterial(mtrl); }

  // Set up lighting states
  D3DUtil_InitLight(light, D3DLIGHT_DIRECTIONAL, -1.0, 1.0, 1.0);
  light.Range:= 1000;
  light.Diffuse:= D3DXColorFromDWord($0082A499);
  D3DXColorScale(light.Diffuse, light.Diffuse, 2);

  m_pd3dDevice.SetLight(0, light);
  m_pd3dDevice.LightEnable(0, True);
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iTrue);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pObject.InvalidateDeviceObjects;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects;
begin
  m_pFont.DeleteDeviceObjects;
  m_pObject.Destroy_;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup;
begin
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pObject);
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(var pCaps: TD3DCaps8; dwBehavior: DWord;
  Format: TD3DFormat): HResult;
begin
  if dwBehavior and D3DCREATE_PUREDEVICE <> 0 then
  begin
    Result:= E_FAIL; // GetTransform doesn't work on PUREDEVICE
    Exit;
  end;

  // If this is a TnL device, make sure it supports directional lights
  if (dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING <> 0) or
     (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING <> 0) then
  begin
    if (pCaps.VertexProcessingCaps and D3DVTXPCAPS_DIRECTIONALLIGHTS = 0) then
    begin
      Result:= E_FAIL;
      Exit;
    end;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Overrrides the main WndProc, so the sample can do custom message
//       handling (e.g. processing mouse, keyboard, or menu commands).
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT;
begin
  case uMsg of
    WM_LBUTTONDOWN: SetCapture(hWnd); // User pressed left mouse button
    WM_LBUTTONUP:   ReleaseCapture;   // The user released the left mouse button
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

procedure CMyD3DApplication.GenerateXFile;
var
  m_NewMesh: ID3DXMesh; // This mesh interface waht we will generate
  i, j: Integer;
  vb: PMyVertex;
  ib: PWord;
  pAdjacencyBuffer: ID3DXBuffer;
begin
  // Note ID3DXMesh are stored internally as list of indexed triangles
  // and (by default) indexes are 16-bit

  // Create mesh object with needed faces (number of indexes) and vertices count
  D3DXCreateMeshFVF(numTriangles, numTriangles*3, D3DXMESH_SYSTEMMEM,
    D3DFVF_VERTEX, m_pd3dDevice, m_NewMesh);

  // Fill Vertex Buffer
  m_NewMesh.LockVertexBuffer(D3DLOCK_NOSYSLOCK, PByte(vb));
  for i:= 0 to numTriangles - 1 do
    for j:= 0 to 2 do
  begin
    vb.p:= D3DXVector3(
             FaceData.Vertices[Face[i, j], 0],
             FaceData.Vertices[Face[i, j], 1],
             FaceData.Vertices[Face[i, j], 2]);
    vb.n:= D3DXVector3(
             FaceData.Normals[Face[i, j+3], 0],
             FaceData.Normals[Face[i, j+3], 1],
             FaceData.Normals[Face[i, j+3], 2]);
    Inc(vb);
  end;
  m_NewMesh.UnlockVertexBuffer;

  // Fill Index Buffer
  m_NewMesh.LockIndexBuffer(D3DLOCK_NOSYSLOCK, PByte(ib));
  for i:= 0 to numTriangles*3 - 1 do
  begin
    ib^:= DWORD(i);
    Inc(ib);
  end;
  m_NewMesh.UnlockIndexBuffer;


  // Let's try to optimize our model
  // To do it we need to have adjacency information about mesh, so generate it
  D3DXCreateBuffer(3 * m_NewMesh.GetNumFaces * SizeOf(DWORD), pAdjacencyBuffer);
  m_NewMesh.GenerateAdjacency(0.001, pAdjacencyBuffer.GetBufferPointer);

  m_NewMesh.OptimizeInplace(
    D3DXMESHOPT_COMPACT or D3DXMESHOPT_ATTRSORT or D3DXMESHOPT_VERTEXCACHE,
    pAdjacencyBuffer.GetBufferPointer, nil, nil, nil);

  // Create X-File in binary format - you can create compressed files (commented out)
  D3DXSaveMeshToX('Face.x', m_NewMesh, nil, nil, 0,
    DWORD(DXFILEFORMAT_BINARY) or DWORD(DXFILEFORMAT_COMPRESSED));
  // Create X-File in text format - they are easy to read
  D3DXSaveMeshToX('FaceText.x', m_NewMesh, nil, nil, 0,
    DWORD(DXFILEFORMAT_TEXT));

  SAFE_RELEASE(m_NewMesh);
end;

end.
